from typing import Union
from energizer.tensor import Tensor
import numpy as np
import numpy.lib.stride_tricks as np_stride_tricks
import mlx.core as mx

from energizer.function import Function
import energizer.derivatives as dv


def max(tensor: Tensor, floor: Union[int, float] = 0) -> Tensor:
    if tensor.device == "gpu":
        return Tensor(
            mx.maximum(floor, tensor.data),
            requires_grad=tensor.requires_grad,
            grad_fn=Function(dv.max_backward, [tensor, floor]),
            device=tensor.device,
        )
    return Tensor(
        np.maximum(floor, tensor.data),
        requires_grad=tensor.requires_grad,
        grad_fn=Function(dv.max_backward, [tensor, floor]),
        device=tensor.device,
    )


def as_strided(
    tensor: Tensor, shape: tuple, strides: tuple, storage_offset: int = 0
) -> "Tensor":
    array = tensor.data

    itemsize = array.itemsize
    byte_strides = tuple(s * itemsize for s in strides)

    if storage_offset > 0:
        flat_array = array.flatten()
        if storage_offset >= len(flat_array):
            raise ValueError(
                f"storage_offset {storage_offset} is out of bounds for array of size {len(flat_array)}"
            )
        offset_array = flat_array[storage_offset:]
        array = offset_array

    if tensor.device == "gpu":
        strided_data = mx.as_strided(
            array, shape=shape, strides=byte_strides, storage_offset=storage_offset
        )
    else:
        strided_data = np_stride_tricks.as_strided(
            array, shape=shape, strides=byte_strides
        )

    return Tensor(
        strided_data,
        requires_grad=tensor.requires_grad,
        grad_fn=Function(
            dv.as_strided_backward, [tensor, shape, strides, storage_offset]
        ),
        device=tensor.device,
    )


def trace(tensor: Tensor) -> Tensor:
    if tensor.device == "gpu":
        result = mx.trace(tensor.data)
    else:
        result = np.trace(tensor.data)
    return Tensor(
        result,
        requires_grad=tensor.requires_grad,
        grad_fn=Function(dv.trace_backward, [tensor]),
        device=tensor.device,
    )


def tanh(tensor: Tensor) -> Tensor:
    if tensor.device == "gpu":
        return Tensor(
            mx.tanh(tensor.data),
            requires_grad=tensor.requires_grad,
            grad_fn=Function(dv.tanh_backward, [tensor]),
            device=tensor.device,
        )
    return Tensor(
        np.tanh(tensor.data),
        requires_grad=tensor.requires_grad,
        grad_fn=Function(dv.tanh_backward, [tensor]),
        device=tensor.device,
    )


def softmax(tensor: Tensor) -> Tensor:
    if tensor.device == "gpu":
        return Tensor(
            mx.softmax(tensor.data),
            requires_grad=tensor.requires_grad,
            grad_fn=Function(dv.softmax_backward, [tensor]),
            device=tensor.device,
        )
    return Tensor(
        np.softmax(tensor.data),
        requires_grad=tensor.requires_grad,
        grad_fn=Function(dv.softmax_backward, [tensor]),
        device=tensor.device,
    )



from energizer.neural_network import Module
from energizer.tensor import Tensor
import numpy as np
import mlx.core as mx
from energizer.function import Function
import energizer.derivatives as dv


class Sigmoid(Module):
    def __init__(self, device: str = "cpu"):
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        if x.device == "gpu":
            y = 1 / (1 + mx.exp(-x.data))
        else:
            y = 1 / (1 + np.exp(-x.data))

        return Tensor(
            y,
            requires_grad=x.requires_grad,
            grad_fn=Function(dv.sigmoid_backward, [x, y]),
            device=x.device,
        )



from energizer.neural_network import Module
from energizer.tensor import Tensor
from energizer.function import Function
import energizer.derivatives as dv
import numpy as np


class Flatten(Module):
    def __init__(self, start_dim: int = 1, end_dim: int = -1):
        super().__init__()
        self.start_dim = start_dim
        self.end_dim = end_dim

    def forward(self, x: Tensor) -> Tensor:
        batch_size = x.data.shape[0]
        if self.start_dim < 0:
            self.start_dim = len(x.data.shape) + self.start_dim
        if self.end_dim < 0:
            self.end_dim = len(x.data.shape) + self.end_dim

        flattened = x.data.reshape(batch_size, -1)
        return Tensor(
            flattened,
            requires_grad=x.requires_grad,
            grad_fn=Function(dv.reshape_backward, [x]) if x.requires_grad else None,
            device=x.device,
        )



from energizer.neural_network import Module
from energizer.tensor import Tensor
from energizer.function import Function
import energizer.derivatives as dv
import numpy as np

try:
    import mlx.core as mx
except ImportError:
    mx = None


class MSELoss(Module):
    def __init__(
        self, size_average: bool = None, reduce: bool = None, reduction: str = "mean"
    ):
        super().__init__()
        self.size_average = size_average
        self.reduce = reduce
        self.reduction = reduction

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        diff = (input - target) ** 2
        if self.reduction == "sum":
            return diff.sum()
        return diff.mean()

    def __call__(self, input: Tensor, target: Tensor) -> Tensor:
        return self.forward(input, target)


class CrossEntropyLoss(Module):
    def __init__(
        self, size_average: bool = None, reduce: bool = None, reduction: str = "mean"
    ):
        super().__init__()
        self.size_average = size_average
        self.reduce = reduce
        self.reduction = reduction

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        logits_np = np.array(input.data).astype(np.float32)
        target_np = np.array(target.data)

        B = logits_np.shape[0]

        shifted = logits_np - logits_np.max(axis=1, keepdims=True)
        log_probs = shifted - np.log(np.exp(shifted).sum(axis=1, keepdims=True))

        if target_np.ndim == 1:
            nll = -log_probs[np.arange(B), target_np.astype(int)]
        else:
            nll = -(target_np.astype(np.float32) * log_probs).sum(axis=1)

        if self.reduction == "sum":
            loss_val = float(nll.sum())
        elif self.reduction == "mean":
            loss_val = float(nll.mean())
        else:
            loss_val = nll

        if isinstance(loss_val, float):
            if mx is not None and input.device == "gpu":
                loss_data = mx.array(loss_val)
            else:
                loss_data = np.array(loss_val, dtype=np.float32)
        else:
            loss_data = loss_val.astype(np.float32)

        return Tensor(
            loss_data,
            requires_grad=input.requires_grad,
            grad_fn=Function(dv.cross_entropy_backward, [input, target]),
            device=input.device,
        )

    def __call__(self, input: Tensor, target: Tensor) -> Tensor:
        return self.forward(input, target)



from energizer.neural_network import Module
from energizer.tensor import Tensor
import numpy as np

class Tanh(Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        return Tensor(np.tanh(x.data), requires_grad=x.requires_grad, device=x.device)


from energizer.neural_network import Module
from energizer.tensor import Tensor
from energizer.function import Function
import energizer.derivatives as dv
import numpy as np

try:
    import mlx.core as mx
except ImportError:
    mx = None


class GELU(Module):
    def __init__(self, device: str = "cpu"):
        super().__init__(device=device)

    def forward(self, x: Tensor) -> Tensor:
        if x.device == "gpu" and mx is not None:
            val = (
                0.5
                * x.data
                * (
                    1.0
                    + mx.tanh(
                        mx.array(np.sqrt(2.0 / np.pi))
                        * (x.data + 0.044715 * mx.power(x.data, 3))
                    )
                )
            )
        else:
            val = (
                0.5
                * x.data
                * (
                    1.0
                    + np.tanh(
                        np.sqrt(2.0 / np.pi) * (x.data + 0.044715 * np.power(x.data, 3))
                    )
                )
            )

        return Tensor(
            val,
            requires_grad=x.requires_grad,
            grad_fn=Function(dv.gelu_backward, [x]),
            device=x.device,
        )



from energizer.neural_network import Module
from energizer.tensor import Tensor
from energizer.function import Function
import energizer.derivatives as dv
import numpy as np


class Reshape(Module):
    def __init__(self, shape: tuple):
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        if not isinstance(x, Tensor):
            raise TypeError(f"Expected Tensor, got {type(x)}")

        new_shape = list(self.shape)

        if -1 in new_shape:
            total_elements = np.prod(x.shape)
            known_elements = np.prod([abs(d) for d in new_shape if d != -1])
            inferred_dim = total_elements // known_elements
            new_shape[new_shape.index(-1)] = inferred_dim

        new_shape = tuple(new_shape)

        if x.device == "gpu":
            import mlx.core as mx

            if isinstance(x.data, mx.array):
                reshaped_data = mx.reshape(x.data, new_shape)
            else:
                reshaped_data = mx.reshape(mx.array(x.data), new_shape)
        else:
            if isinstance(x.data, np.ndarray):
                reshaped_data = x.data.reshape(new_shape)
            else:
                reshaped_data = np.array(x.data).reshape(new_shape)

        return Tensor(
            reshaped_data,
            requires_grad=x.requires_grad,
            grad_fn=Function(dv.reshape_backward, [x]) if x.requires_grad else None,
            device=x.device,
        )



import numpy as np
import mlx.core as mx
from typing import Union, Optional

import energizer.derivatives as dv
from .function import Function


class Tensor:
    """
    A tensor is a multi-dimensional array.
    """

    def __init__(
        self,
        data,
        requires_grad: bool = False,
        grad_fn: Optional["Function"] = None,
        device: str = "cpu",
    ):
        self.device = device
        self.requires_grad = requires_grad
        self.grad_fn = grad_fn
        self.grad = None

        if self.device == "cpu":
            self._init_cpu(data)
        elif self.device == "gpu":
            self._init_mlx(data)

    def __str__(self):
        # Simple debugger-style representation showing device and backend
        backend = "mlx" if isinstance(self.data, mx.array) else "numpy"
        data_str = np.array2string(
            np.array(self.data),
            separator=", ",
            precision=4,
            suppress_small=False,
            floatmode="maxprec_equal",
        )
        return (
            f"tensor({data_str}, "
            f"device={self.device}, "
            f"backend={backend}, "
            f"requires_grad={self.requires_grad})"
        )

    def _init_cpu(self, data):
        if isinstance(data, np.ndarray):
            self.data = data
        elif isinstance(data, mx.array):
            self.data = np.array(data)
        elif isinstance(data, Tensor):
            if data.device == "gpu":
                self.data = np.array(data.data)
            else:
                self.data = data.data.copy()
        else:
            self.data = np.array(data)

        self.shape = self.data.shape

    def _init_mlx(self, data):
        if isinstance(data, mx.array):
            self.data = data
        elif isinstance(data, np.ndarray):
            self.data = mx.array(data)
        elif isinstance(data, Tensor):
            if data.device == "gpu":
                self.data = data.data
            else:
                self.data = mx.array(data.data)
        else:
            self.data = mx.array(data)

        self.shape = self.data.shape
        self.device = "gpu"

    def cpu(self) -> "Tensor":
        if self.device == "cpu":
            return self
        return Tensor(
            np.array(self.data), requires_grad=self.requires_grad, device="cpu"
        )

    def gpu(self) -> "Tensor":
        if self.device == "gpu":
            return self
        return Tensor(
            mx.array(self.data), requires_grad=self.requires_grad, device="gpu"
        )

    def to(self, device: str) -> "Tensor":
        if self.device == device:
            return self
        if device == "cpu":
            return self.cpu()
        elif device == "gpu":
            return self.gpu()
        else:
            raise ValueError(f"Invalid device: {device}")

    def numpy(self) -> np.ndarray:
        if isinstance(self.data, mx.array):
            return np.array(self.data)
        return self.data.copy()

    def mlx(self) -> mx.array:
        if isinstance(self.data, mx.array):
            return self.data
        return mx.array(self.data)

    def _get_other_data(self, other: Union[int, float, "Tensor"]):
        if isinstance(other, (int, float)):
            return other
        elif isinstance(other, Tensor):
            if self.device == "cpu":
                return other.cpu().data
            elif self.device == "gpu":
                return other.gpu().data
        # Handle MLX arrays and numpy arrays directly
        elif isinstance(other, mx.array):
            if self.device == "cpu":
                return np.array(other)
            else:
                return other
        elif isinstance(other, np.ndarray):
            if self.device == "gpu":
                return mx.array(other)
            else:
                return other
        # Fallback: try to access .data if it exists
        elif hasattr(other, "data"):
            return other.data
        else:
            return other

    def requires_grad_(self, requires_grad: bool = True):
        self.requires_grad = requires_grad
        return self

    def __mul__(self, other: Union[int, float, "Tensor"]) -> "Tensor":
        other_data = self._get_other_data(other)
        return Tensor(
            self.data * other_data,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.mul_backward, [self, other]),
            device=self.device,
        )

    def __rmul__(self, other: Union[int, float]) -> "Tensor":
        return Tensor(
            other * self.data,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.mul_backward, [self, other]),
            device=self.device,
        )

    def __add__(self, other: Union[int, float, "Tensor"]) -> "Tensor":
        other_data = self._get_other_data(other)
        return Tensor(
            self.data + other_data,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.add_backward, [self, other]),
            device=self.device,
        )

    def __radd__(self, other: Union[int, float, "Tensor"]) -> "Tensor":
        return self.__add__(other)

    def __sub__(self, other: Union[int, float, "Tensor"]) -> "Tensor":
        other_data = self._get_other_data(other)
        return Tensor(
            self.data - other_data,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.sub_backward, [self, other]),
            device=self.device,
        )

    def __neg__(self) -> "Tensor":
        return Tensor(
            (
                isinstance(self.data, (int, float)) - self.data
                if isinstance(self.data, (int, float))
                else -self.data
            ),
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.neg_backward, [self]),
            device=self.device,
        )

    def __truediv__(self, other: Union[int, float, "Tensor"]) -> "Tensor":
        other_data = self._get_other_data(other)
        return Tensor(
            self.data / other_data,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.truediv_backward, [self, other]),
            device=self.device,
        )

    def __rtruediv__(self, other: Union[int, float]) -> "Tensor":
        other = self._get_other_data(other)
        return Tensor(
            other / self.data,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.truediv_backward, [other, self]),
            device=self.device,
        )

    def __matmul__(self, other: "Tensor") -> "Tensor":
        other_data = self._get_other_data(other)
        return Tensor(
            self.data @ other_data,
            requires_grad=self.requires_grad or other.requires_grad,
            grad_fn=Function(dv.matmul_backward, [self, other]),
            device=self.device,
        )

    def __pow__(self, other: Union[int, float]) -> "Tensor":
        return Tensor(
            self.data**other,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.pow_backward, [self, other]),
            device=self.device,
        )

    def __rmatmul__(self, other: "Tensor") -> "Tensor":
        return self.__matmul__(other)

    def __getitem__(self, key):
        return Tensor(
            self.data[key],
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.getitem_backward, [self, key]),
            device=self.device,
        )

    def __setitem__(self, key, value: Union[int, float, "Tensor"]):
        if isinstance(value, Tensor):
            value_data = value.data
        else:
            value_data = value

        if isinstance(self.data, mx.array):
            temp_np = np.array(self.data)
            temp_np[key] = value_data
            new_data = mx.array(temp_np)
        else:
            new_data = self.data.copy()
            new_data[key] = value_data

        if self.requires_grad:
            self.grad_fn = Function(dv.setitem_backward, [self, key, value])

        self.data = new_data

    @property
    def T(self) -> "Tensor":
        return Tensor(
            self.data.T,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.transpose_backward, [self]),
            device=self.device,
        )

    @property
    def size(self):
        return self.data.shape

    def item(self) -> Union[int, float]:
        if self.data.size != 1:
            raise ValueError("Tensor must be a single element")

        scalar_value = self.data.item()
        if self.requires_grad:
            self.grad_fn = Tensor(
                scalar_value,
                requires_grad=self.requires_grad,
                grad_fn=Function(dv.item_backward, [self]),
                device=self.device,
            )
        return scalar_value

    def copy(self) -> "Tensor":
        return Tensor(
            self.data.copy(), requires_grad=self.requires_grad, grad_fn=self.grad_fn
        )

    def sum(self, dim: Optional[int] = None, axis: Optional[int] = None) -> "Tensor":
        if axis is not None:
            dim = axis
        if isinstance(self.data, mx.array):
            result = mx.sum(self.data, axis=dim)
        else:
            result = np.sum(self.data, axis=dim)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.sum_backward, [self, dim]),
            device=self.device,
        )

    def mean(self, dim: Optional[int] = None, axis: Optional[int] = None) -> "Tensor":
        if axis is not None:
            dim = axis
        if isinstance(self.data, mx.array):
            result = mx.mean(self.data, axis=dim)
        else:
            result = np.mean(self.data, axis=dim)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.mean_backward, [self, dim]),
            device=self.device,
        )

    def squeeze(self, dim: Optional[int] = None, axis: Optional[int] = None) -> "Tensor":
        if axis is not None:
            dim = axis
        if isinstance(self.data, mx.array):
            result = self.data.squeeze(axis=dim)
        else:
            result = self.data.squeeze(axis=dim)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.squeeze_backward, [self, dim]),
            device=self.device,
        )

    def exp(self) -> "Tensor":
        if isinstance(self.data, mx.array):
            result = self.data.exp()
        else:
            result = np.exp(self.data)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.exp_backward, [self]),
            device=self.device,
        )

    def log(self) -> "Tensor":
        if isinstance(self.data, mx.array):
            result = self.data.log()
        else:
            result = np.log(self.data)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.log_backward, [self]),
            device=self.device,
        )

    def clamp(self, min: Union[int, float], max: Union[int, float]) -> "Tensor":
        if isinstance(self.data, mx.array):
            result = mx.clip(self.data, min, max)
        else:
            result = np.clip(self.data, min, max)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.clamp_backward, [self, min, max]),
            device=self.device,
        )

    def minimum(self, other: Union[int, float, "Tensor"]) -> "Tensor":
        other_data = self._get_other_data(other)
        if isinstance(self.data, mx.array):
            result = mx.minimum(self.data, other_data)
        else:
            result = np.minimum(self.data, other_data)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.minimum_backward, [self, other]),
            device=self.device,
        )

    def maximum(self, other: Union[int, float, "Tensor"]) -> "Tensor":
        other_data = self._get_other_data(other)
        if isinstance(self.data, mx.array):
            result = mx.maximum(self.data, other_data)
        else:
            result = np.maximum(self.data, other_data)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.maximum_backward, [self, other]),
            device=self.device,
        )

    def reshape(self, shape: tuple) -> "Tensor":
        if isinstance(self.data, mx.array):
            result = self.data.reshape(shape)
        else:
            result = self.data.reshape(shape)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.reshape_backward, [self, shape]),
            device=self.device,
        )

    def view(self, shape: tuple) -> "Tensor":
        return self.reshape(shape)

    def transpose(self, dim0: int, dim1: int) -> "Tensor":
        if self.device == "gpu":
            result = mx.swapaxes(self.data, dim0, dim1)
        else:
            result = np.swapaxes(self.data, dim0, dim1)
        return Tensor(
            result,
            requires_grad=self.requires_grad,
            grad_fn=Function(dv.transpose_backward, [self, dim0, dim1]),
            device=self.device,
        )

    @staticmethod
    def randn(*args, device: str = "cpu", **kwargs):
        if device == "gpu":
            data = mx.random.normal(shape=args, **kwargs)
        else:
            data = np.random.randn(*args, **kwargs)

        return Tensor(data, requires_grad=False, device=device)

    @staticmethod
    def zeros(shape, device: str = "cpu"):
        if device == "gpu":
            data = mx.zeros(shape)
        else:
            data = np.zeros(shape)
        return Tensor(data, requires_grad=False, device=device)

    @staticmethod
    def ones(shape, device: str = "cpu"):
        if device == "gpu":
            data = mx.ones(shape)
        else:
            data = np.ones(shape)
        return Tensor(data, requires_grad=False, device=device)

    def backward(self, grad_outputs=None):
        if grad_outputs is None:
            if isinstance(self.data, mx.array):
                grad_outputs = Tensor(mx.ones_like(self.data), device=self.device)
            else:
                grad_outputs = Tensor(np.ones_like(self.data), device=self.device)
        else:
            # Convert numpy/mx array to Tensor if needed
            if not isinstance(grad_outputs, Tensor):
                grad_outputs = Tensor(grad_outputs, device=self.device)
            # Ensure grad_outputs is on the same device as self
            elif self.device != grad_outputs.device:
                grad_outputs = grad_outputs.to(self.device)

        if self.requires_grad and self.grad_fn is not None:
            self.grad_fn.backward([grad_outputs])


def tensor(data, requires_grad=False, device="cpu"):
    return Tensor(data, requires_grad=requires_grad, device=device)
